@@ -37,13 +37,19 @@ from .version import __version__
 from .util import NONE_SENTINEL, Sentinel, extract_service_endpoint
 missing = Sentinel("Missing")
 APPEND_USER_AGENT_ENV_VAR_NAME = "OCI_SDK_APPEND_USER_AGENT"
+PROPAGATION_ENABLED_ENV_VAR_NAME = "PROPAGATION_ENABLED"
+OPC_INCOMING_REQUEST_ID_ENV_VAR_NAME = "OPC_INCOMING_REQUEST_ID"
 OCI_REALM_SPECIFIC_SERVICE_ENDPOINT_TEMPLATE_ENABLED = "OCI_REALM_SPECIFIC_SERVICE_ENDPOINT_TEMPLATE_ENABLED"
 APPEND_USER_AGENT = os.environ.get(APPEND_USER_AGENT_ENV_VAR_NAME)
+PROPAGATION_ENABLED = False
 USER_INFO = "Oracle-PythonSDK/{}".format(__version__)
 
 DICT_VALUE_TYPE_REGEX = re.compile(r'dict\(str, (.+?)\)$')  # noqa: W605
 LIST_ITEM_TYPE_REGEX = re.compile(r'list\[(.+?)\]$')  # noqa: W605
 TROUBLESHOOT_URL = 'https://docs.oracle.com/en-us/iaas/Content/API/Concepts/sdk_troubleshooting.htm'
+OCI_DUAL_STACK_ENDPOINT_ENABLED_ENV_VAR = "OCI_DUAL_STACK_ENDPOINT_ENABLED"
+PATTERN_FOR_ENDPOINT_TEMPLATE_OPTIONS = re.compile(r"\{\w*\?(\w*.|\s*)\:(\w*.|\s*)\}*")
+DUAL_STACK_OPTION = "{dualStack"
 
 # Expect header is enabled by default
 enable_expect_header = True
@@ -279,10 +285,15 @@ class BaseClient(object):
         # clients we require an endpoint
         self.regional_client = kwargs.get('regional_client', True)
 
+        self.custom_opc_request_id = None
+        self.PROPAGATION_ENABLED = os.environ.get(PROPAGATION_ENABLED_ENV_VAR_NAME)
+        self.PROPAGATION_ENABLED = False if self.PROPAGATION_ENABLED is None else self.get_bool_env_var(self.PROPAGATION_ENABLED)
         self._endpoint = None
         self._base_path = kwargs.get('base_path')
         self.service_endpoint_template = kwargs.get('service_endpoint_template')
         self.service_endpoint_template_per_realm = kwargs.get('service_endpoint_template_per_realm')
+        self.client_level_dualstack_endpoints_enabled = kwargs.get('client_level_dualstack_endpoints_enabled')
+        self.service_uses_dualstack_endpoints_by_default = kwargs.get("service_uses_dualstack_endpoints_by_default", False)
         self.endpoint_service_name = kwargs.get('endpoint_service_name')
 
         # By default self._allow_control_chars will be None. The user would need to explicitly set it to True or False
@@ -292,6 +303,7 @@ class BaseClient(object):
 
         self._client_level_realm_specific_endpoint_template_enabled = kwargs.get('client_level_realm_specific_endpoint_template_enabled')  # default this to None as it should be an opt-in feature
 
+        self.service = service
         if self.regional_client:
             if kwargs.get('service_endpoint'):
                 self.endpoint = kwargs.get('service_endpoint')
@@ -315,7 +327,6 @@ class BaseClient(object):
                 raise exceptions.MissingEndpointForNonRegionalServiceClientError('An endpoint must be provided for a non-regional service client')
             self.endpoint = kwargs.get('service_endpoint')
 
-        self.service = service
         self.complex_type_mappings = type_mapping
         self.type_mappings = merge_type_mappings(self.primitive_type_map, type_mapping)
         self.session = requests.Session()
@@ -366,6 +377,42 @@ class BaseClient(object):
 
         return service_endpoint_template
 
+    def is_dual_stack_enabled(self):
+        """
+        Returns a boolean for whether dual stack endpoints are enabled or not
+        The hierarchy is:
+        1. Client level setting
+        2. Environment level setting
+        3. Service level setting
+        """
+        if self.client_level_dualstack_endpoints_enabled is not None:
+            return self.client_level_dualstack_endpoints_enabled
+        dual_stack_endpoints_enabled_from_env_var = os.environ.get(OCI_DUAL_STACK_ENDPOINT_ENABLED_ENV_VAR)
+        if dual_stack_endpoints_enabled_from_env_var is not None:
+            return dual_stack_endpoints_enabled_from_env_var.lower() == 'true'
+        return self.service_uses_dualstack_endpoints_by_default
+
+    def update_endpoint_template_for_options(self):
+        pattern = PATTERN_FOR_ENDPOINT_TEMPLATE_OPTIONS
+        endpoint = self.endpoint
+        matchers = re.finditer(pattern, endpoint)
+
+        for matcher in matchers:
+            option = matcher.group()
+            option_enabled_param = option[option.index('?') + 1:option.index(':')]
+            option_disabled_param = option[option.index(':') + 1:-1]
+            # Dual stack option
+            if DUAL_STACK_OPTION in option:
+                if self.is_dual_stack_enabled():
+                    # TODO replacing endpoint options should be factored out in the outer loop when dealing with multiple options
+                    endpoint = endpoint.replace(option, option_enabled_param)
+                else:
+                    endpoint = endpoint.replace(option, option_disabled_param)
+            else:
+                endpoint = endpoint.replace(option, '')
+
+        return endpoint
+
     @property
     def endpoint(self):
         return self._endpoint
@@ -438,7 +485,8 @@ class BaseClient(object):
                 isinstance(self.signer, signers.EphemeralResourcePrincipalSigner) or
                 isinstance(self.signer, signers.EphemeralResourcePrincipalV21Signer) or
                 isinstance(self.signer, signers.NestedResourcePrincipals) or
-                (isinstance(self.signer, signers.OkeWorkloadIdentityResourcePrincipalSigner) and oke_workload_refresh_enabled)):
+                (isinstance(self.signer, signers.OkeWorkloadIdentityResourcePrincipalSigner) and oke_workload_refresh_enabled) or
+                isinstance(self.signer, signers.OauthExchangeTokenSigner)):
             return True
         else:
             return False
@@ -492,8 +540,24 @@ class BaseClient(object):
         header_params[constants.HEADER_CLIENT_INFO] = USER_INFO
         header_params[constants.HEADER_USER_AGENT] = self.user_agent
 
-        if header_params.get(constants.HEADER_REQUEST_ID, missing) is missing:
+        self.get_downstream_request_id()
+
+        # Set custom opc-request-id header, if specified in the client
+        if self.custom_opc_request_id is None and self.PROPAGATION_ENABLED in [False, "False", None, "None"] and header_params.get(constants.HEADER_REQUEST_ID, missing) is missing:
             header_params[constants.HEADER_REQUEST_ID] = self.build_request_id()
+            self.logger.debug(f"No propagation: {str(header_params[constants.HEADER_REQUEST_ID])}")
+        elif self.custom_opc_request_id is None and self.PROPAGATION_ENABLED in [False, "False", None, "None"] and header_params.get(constants.HEADER_REQUEST_ID):
+            self.logger.debug(f"No propagation: {str(header_params[constants.HEADER_REQUEST_ID])}")
+        elif self.PROPAGATION_ENABLED in [False, "False", None, "None"] and header_params.get(constants.HEADER_REQUEST_ID, missing) is missing:
+            header_params[constants.HEADER_REQUEST_ID] = self.build_request_id()
+            self.logger.debug(f"Propagation disabled with no other values received from SDK or CLI: {str(header_params[constants.HEADER_REQUEST_ID])}")
+        elif self.PROPAGATION_ENABLED in [True, "True"] and header_params.get(constants.HEADER_REQUEST_ID, missing) is missing and self.custom_opc_request_id is None:
+            clientId = self.build_request_id()
+            header_params[constants.HEADER_REQUEST_ID] = self.use_custom_opc_request_id(clientId)
+            self.logger.debug(f"Propagation enabled with no other values received:{self.PROPAGATION_ENABLED} and {str(header_params[constants.HEADER_REQUEST_ID])} and {self.custom_opc_request_id}")
+        elif self.PROPAGATION_ENABLED in [True, "True"] and header_params.get(constants.HEADER_REQUEST_ID, missing) is missing and self.custom_opc_request_id is not None:
+            header_params[constants.HEADER_REQUEST_ID] = self.custom_opc_request_id
+            self.logger.debug(f"Propagation enabled with no other values received from SDK but CLI: {str(header_params[constants.HEADER_REQUEST_ID])}")
 
         # This allows for testing with "fake" database resources.
         opc_host_serial = os.environ.get('OCI_DB_OPC_HOST_SERIAL')
@@ -550,6 +614,53 @@ class BaseClient(object):
             self.logger.debug('time elapsed for request: {}'.format(str(end - start)))
             return response
 
+    def get_bool_env_var(envVar: str, default=False) -> bool:
+        if envVar is None:
+            return False
+        envVar = str(envVar).strip().lower()
+        if envVar in ('True', 'true', 'TRUE'):
+            return True
+        elif envVar in ('False', 'false', 'FALSE'):
+            return False
+        return default
+
+    def use_default_opc_request_id(self):
+        self.PROPAGATION_ENABLED = False
+        self.custom_opc_request_id = None
+        if os.environ.get(PROPAGATION_ENABLED_ENV_VAR_NAME):
+            os.environ.pop(PROPAGATION_ENABLED_ENV_VAR_NAME)
+        if os.environ.get(OPC_INCOMING_REQUEST_ID_ENV_VAR_NAME):
+            os.environ.pop(OPC_INCOMING_REQUEST_ID_ENV_VAR_NAME)
+
+    def get_downstream_request_id(self):
+        request_id = os.environ.get(OPC_INCOMING_REQUEST_ID_ENV_VAR_NAME)
+        if self.PROPAGATION_ENABLED and request_id:
+            self.logger.debug(f"Downstream requestID: {str(request_id)}")
+            self.custom_opc_request_id = request_id
+            return request_id
+
+    def use_custom_opc_request_id(self, rid: str = None):
+        self.custom_opc_request_id = None
+        if self.PROPAGATION_ENABLED:
+            rid = self.get_downstream_request_id()
+        if rid is not None and rid != "":
+            self.PROPAGATION_ENABLED = True
+        if self.PROPAGATION_ENABLED:
+            segments = rid.split('/')
+            if len(segments) == 1:
+                stackId = self.build_request_id()
+                self.custom_opc_request_id = segments[0] + "/" + stackId
+                self.logger.debug(f"Generated stackId: {str(stackId)}")
+            elif len(segments) >= 2:
+                if len(segments[1]) == 0:
+                    stackId = self.build_request_id()
+                    self.logger.debug(f"Generated stackId: {str(stackId)}")
+                else:
+                    stackId = segments[1]
+                self.custom_opc_request_id = segments[0] + "/" + stackId
+                self.logger.debug(f"Truncated request ID to: {str(self.custom_opc_request_id)}")
+        return self.custom_opc_request_id
+
     def map_service_params_to_values(self, service_params_url, path_params, query_params, required_arguments):
         service_params_map = {}
         service_params = set(filter(None, ("".join(service_params_url.replace(".", "").split("{"))).split("}")))
@@ -577,7 +688,7 @@ class BaseClient(object):
         return service_params_map
 
     def handle_service_params_in_endpoint(self, path_params, query_params, required_arguments):
-        endpoint = self.endpoint
+        endpoint = self.update_endpoint_template_for_options()
         start_idx = len("https://")
         if endpoint[start_idx] == "{":  # If the character after https:// is a "{", we have service params
             end_idx = endpoint.rfind("}")
@@ -701,6 +812,9 @@ class BaseClient(object):
                 self.logger.debug(utc_now() + 'time elapsed for request {}: {}'.format(request.header_params[constants.HEADER_REQUEST_ID], str(end - start)))
             if response and hasattr(response, 'elapsed'):
                 self.logger.debug(utc_now() + "time elapsed in response: " + str(response.elapsed))
+            if self.PROPAGATION_ENABLED in [True, "True"] and response.headers[constants.HEADER_REQUEST_ID]:
+                os.environ[OPC_INCOMING_REQUEST_ID_ENV_VAR_NAME] = response.headers[constants.HEADER_REQUEST_ID]
+                self.logger.debug(f"Response opc-request-id: {response.headers[constants.HEADER_REQUEST_ID]}")
         except requests.exceptions.ConnectTimeout as e:
             if not e.args:
                 e.args = ('',)
@@ -766,6 +880,27 @@ class BaseClient(object):
     def build_request_id(self):
         return str(uuid.uuid4()).replace('-', '').upper()
 
+    def is_valid_opc_request_id(self, rid: str):
+        """
+        Validate the OPC request ID.
+        Args:
+        rid (str): The OPC request ID to be validated.
+        Raises:
+        ValueError: If the OPC request ID is invalid.
+        """
+        self.logger.info(f"custom opc-request-id: {rid}")
+        if rid is None or rid == "":
+            raise ValueError("custom opc-request-id cannot be empty")
+        segments = rid.split("/")
+        if len(segments) > 3:
+            raise ValueError("custom opc-request-id cannot contain more than 3 segments")
+        # pattern = re.compile("^[a-zA-Z0-9_-]{0,32}$")
+        # TODO Change once regex is confirmed to be removed as objectstoarge request id has special character
+        pattern = re.compile("^[a-zA-Z0-9_:-;]{0,32}$")
+        for segment in segments:
+            if not pattern.match(segment):
+                raise ValueError("custom opc-request-id segments must contain only ASCII alphanumerics plus underscore and dash, and be at most 32 characters")
+
     def add_opc_retry_token_if_needed(self, header_params, retry_token_length=30):
         if 'opc-retry-token' not in header_params:
             header_params['opc-retry-token'] = ''.join(random.SystemRandom().choice(string.ascii_letters + string.digits) for _ in range(retry_token_length))
