@@ -46,8 +46,7 @@ REGIONS_SHORT_NAMES = {
     'onm': 'ap-delhi-1',
     'hsg': 'ap-batam-1',
     'orf': 'eu-madrid-3',
-    'des': 'ap-chennai-1',
-    'vkz': 'us-columbus-1',
+    'nrq': 'eu-turin-1',
     'lfi': 'us-langley-1',
     'luf': 'us-luke-1',
     'ric': 'us-gov-ashburn-1',
@@ -130,8 +129,7 @@ REGION_REALMS = {
     'ap-delhi-1': 'oc1',
     'ap-batam-1': 'oc1',
     'eu-madrid-3': 'oc1',
-    'ap-chennai-1': 'oc1',
-    'us-columbus-1': 'oc1',
+    'eu-turin-1': 'oc1',
 
     'us-langley-1': 'oc2',
     'us-luke-1': 'oc2',
@@ -253,8 +251,7 @@ REGIONS = [
     'ap-delhi-1',
     'ap-batam-1',
     'eu-madrid-3',
-    'ap-chennai-1',
-    'us-columbus-1',
+    'eu-turin-1',
     'us-langley-1',
     'us-luke-1',
     'us-gov-ashburn-1',
